package org.bitbucket.bradleysmithllc.etlagent.informatica.handler;

import com.fasterxml.jackson.databind.JsonNode;
import com.google.common.cache.*;
import com.google.gson.GsonBuilder;
import org.apache.commons.pool.ObjectPool;
import org.apache.commons.pool.PoolableObjectFactory;
import org.apache.commons.pool.impl.GenericObjectPool;
import org.bitbucket.bradleysmithllc.etlagent.dto.GenericResponseDTO;
import org.bitbucket.bradleysmithllc.etlagent.informatica.InformaticaRequestHandler;
import org.bitbucket.bradleysmithllc.etlagent.informatica.provider.*;
import org.bitbucket.bradleysmithllc.etlunit.PrintWriterLog;
import org.bitbucket.bradleysmithllc.etlunit.RuntimeSupport;
import org.bitbucket.bradleysmithllc.etlunit.feature.FeatureOperation;

import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;

public abstract class AbstractHandler<T, J extends GenericResponseDTO> implements InformaticaRequestHandler<T, J>
{
	protected RuntimeSupport runtimeSupport;

	private static class InformaticaExecutionService
	{
		private final InformaticaIntegrationService informaticaIntegrationService;
		private final InformaticaWebServicesHub informaticaWebServicesHub;

		private InformaticaExecutionService(InformaticaIntegrationService informaticaIntegrationService, InformaticaWebServicesHub informaticaWebServicesHub) {
			this.informaticaIntegrationService = informaticaIntegrationService;
			this.informaticaWebServicesHub = informaticaWebServicesHub;
		}
	}

	private static final LoadingCache<InformaticaRepository, ObjectPool<InformaticaRepositoryClient>> repositoryClientCaches = CacheBuilder.newBuilder()
			.expireAfterAccess(60, TimeUnit.MINUTES)
			.removalListener(new RemovalListener<InformaticaRepository, ObjectPool<InformaticaRepositoryClient>>() {
				@Override
				public void onRemoval(RemovalNotification<InformaticaRepository, ObjectPool<InformaticaRepositoryClient>> objectObjectRemovalNotification) {
					if (objectObjectRemovalNotification.wasEvicted())
					{
						cacheLog("Disposing Repository service pool for " + objectObjectRemovalNotification.getKey() + " with id " + System.identityHashCode(objectObjectRemovalNotification.getValue()));
						try {
							objectObjectRemovalNotification.getValue().clear();
						} catch (Exception e) {
							cacheLog("Pool could not be disposed " + e.toString());
						}
					}
					else
					{
						cacheLog("Reusing Repository service pool id " + System.identityHashCode(objectObjectRemovalNotification.getValue()));
					}
				}
			})
			.build(
					new CacheLoader<InformaticaRepository, ObjectPool<InformaticaRepositoryClient>>() {
						public ObjectPool<InformaticaRepositoryClient> load(final InformaticaRepository key) {
							GenericObjectPool<InformaticaRepositoryClient> informaticaRepositoryClientGenericObjectPool = new GenericObjectPool<InformaticaRepositoryClient>(new PoolableObjectFactory<InformaticaRepositoryClient>() {
								@Override
								public InformaticaRepositoryClient makeObject() throws Exception {
									cacheLog("Allocating repository service client for " + key);
									InformaticaRepositoryClient informaticaRepositoryClient = null;

									try {
										informaticaRepositoryClient = key.newInformaticaRepositoryClient();
									} finally {
										cacheLog("Allocated repository service client for " + key + " with id " + System.identityHashCode(informaticaRepositoryClient));
									}

									return informaticaRepositoryClient;
								}

								@Override
								public void destroyObject(InformaticaRepositoryClient informaticaRepositoryClient) throws Exception {
									informaticaRepositoryClient.dispose();
								}

								@Override
								public boolean validateObject(InformaticaRepositoryClient informaticaRepositoryClient) {
									return true;
								}

								@Override
								public void activateObject(InformaticaRepositoryClient informaticaRepositoryClient) throws Exception {
								}

								@Override
								public void passivateObject(InformaticaRepositoryClient informaticaRepositoryClient) throws Exception {
								}
							});

							informaticaRepositoryClientGenericObjectPool.setMaxActive(10);
							informaticaRepositoryClientGenericObjectPool.setMaxIdle(10);
							informaticaRepositoryClientGenericObjectPool.setTimeBetweenEvictionRunsMillis(60000L);

							return informaticaRepositoryClientGenericObjectPool;
						}
					});

	private static void cacheLog(String println) {
		System.out.println("[CACHE][thread:" + Thread.currentThread().hashCode() + "] " + println);
	}

	protected static InformaticaRepositoryClient newRepositoryClient(InformaticaRepository repository) throws Exception {
		cacheLog("Handler waiting for client");

		InformaticaRepositoryClient informaticaRepositoryClient = null;

		try
		{
			cacheLog("Handler receiving client [" + Thread.currentThread().hashCode() + "]");
			synchronized(repositoryClientCaches)
			{
				informaticaRepositoryClient = repositoryClientCaches.get(repository).borrowObject();
			}
		}
		finally
		{
			cacheLog("Handler received client [" + Thread.currentThread().hashCode() + "]:[" + System.identityHashCode(informaticaRepositoryClient) + "]");
		}

		return informaticaRepositoryClient;
	}

	private static final LoadingCache<InformaticaExecutionService, ObjectPool<InformaticaExecutionClient>> integrationServiceClientCache = CacheBuilder.newBuilder()
			.expireAfterAccess(60, TimeUnit.MINUTES)
			.removalListener(new RemovalListener<InformaticaExecutionService, ObjectPool<InformaticaExecutionClient>>() {
				@Override
				public void onRemoval(RemovalNotification<InformaticaExecutionService, ObjectPool<InformaticaExecutionClient>> objectObjectRemovalNotification) {
					if (objectObjectRemovalNotification.wasEvicted())
					{
						cacheLog("Disposing Integration service pool for " + objectObjectRemovalNotification.getKey() + " with id " + System.identityHashCode(objectObjectRemovalNotification.getValue()));
						try {
							objectObjectRemovalNotification.getValue().clear();
						} catch (Exception e) {
							cacheLog("Pool could not be disposed " + e.toString());
						}
					}
					else
					{
						cacheLog("Reusing Integration service pool id " + System.identityHashCode(objectObjectRemovalNotification.getValue()));
					}
				}
			})
			.build(
					new CacheLoader<InformaticaExecutionService, ObjectPool<InformaticaExecutionClient>>() {
						public ObjectPool<InformaticaExecutionClient> load(final InformaticaExecutionService key) {
							return new GenericObjectPool<InformaticaExecutionClient>(new PoolableObjectFactory<InformaticaExecutionClient>() {
								@Override
								public InformaticaExecutionClient makeObject() throws Exception {
									InformaticaExecutionClient informaticaIntegrationServiceClient = null;

									// check for a web services hub, and use it if present
									if (key.informaticaWebServicesHub != null)
									{
										cacheLog("Allocating Web services hub client for " + key.informaticaIntegrationService);
										informaticaIntegrationServiceClient = key.informaticaWebServicesHub.newInformaticaWebServicesHubClient(key.informaticaIntegrationService);
										cacheLog("Allocated Web services hub client for " + key.informaticaIntegrationService + " with id " + System.identityHashCode(informaticaIntegrationServiceClient));
									}
									else
									{
										cacheLog("Allocating Integration service client for " + key.informaticaIntegrationService);
										informaticaIntegrationServiceClient = key.informaticaIntegrationService.newInformaticaIntegrationServiceClient();
										cacheLog("Allocated Integration service client for " + key.informaticaIntegrationService + " with id " + System.identityHashCode(informaticaIntegrationServiceClient));
									}

									return informaticaIntegrationServiceClient;
								}

								@Override
								public void destroyObject(InformaticaExecutionClient obj) throws Exception {
									//obj.dispose();
								}

								@Override
								public boolean validateObject(InformaticaExecutionClient obj) {
									return true;
								}

								@Override
								public void activateObject(InformaticaExecutionClient obj) throws Exception {
								}

								@Override
								public void passivateObject(InformaticaExecutionClient obj) throws Exception {
								}
							});
						}
					});

	protected static InformaticaExecutionClient newIntegrationClient(InformaticaExecutionService service) throws Exception {
		synchronized(integrationServiceClientCache)
		{
			return integrationServiceClientCache.get(service).borrowObject();
		}
	}

	static
	{
		Thread thread = new Thread(new Runnable() {
			@Override
			public void run() {
				// clean up both caches
				while (true)
				{
					repositoryClientCaches.cleanUp();
					integrationServiceClientCache.cleanUp();

					try {
						Thread.sleep(60000L);
					} catch (InterruptedException e) {
						e.printStackTrace();
					}
				}
			}
		});

		thread.setDaemon(true);

		thread.start();
	}

	@Override
	public GsonBuilder configureBuilder(GsonBuilder builder) {
		return builder;
	}

	protected PrintWriterLog log = new PrintWriterLog();

	private FeatureOperation featureOperation;
	protected InformaticaConfiguration informaticaConfiguration;

	@Override
	public void setInformaticaConfiguration(InformaticaConfiguration informaticaConfiguration) {
		this.informaticaConfiguration = informaticaConfiguration;
	}

	@Override
	public void receiveRuntimeSupport(RuntimeSupport runtimeSupport) {
		this.runtimeSupport = runtimeSupport;
	}

	@Override
	public final void setFeatureOperation(FeatureOperation featureOperation) {
		this.featureOperation = featureOperation;
	}

	@Override
	public final JsonNode getValidator() {
		return featureOperation.getSignatures().get(0).getValidatorNode();
	}

	public interface RepositoryVisitor
	{
		public void useClient(InformaticaRepositoryClient client, RuntimeSupport runtimeSupport) throws Exception;
	}

	protected void useRepository(InformaticaRepository repository, RepositoryVisitor visitor) throws Exception {

		InformaticaRepositoryClient informaticaRepositoryClient = newRepositoryClient(repository);

		try
		{
			visitor.useClient(informaticaRepositoryClient, runtimeSupport);
		} finally
		{
			cacheLog("Returning Repository service client id " + System.identityHashCode(informaticaRepositoryClient));
			synchronized(repositoryClientCaches)
			{
				repositoryClientCaches.get(repository).returnObject(informaticaRepositoryClient);
			}
		}
	}

	public interface IntegrationServiceVisitor
	{
		public void useClient(InformaticaExecutionClient client, RuntimeSupport runtimeSupport) throws Exception;
	}

	protected void useExecutionService(InformaticaRepository repository, InformaticaIntegrationService intsvc, InformaticaWebServicesHub informaticaWebServicesHub, IntegrationServiceVisitor visitor) throws Exception {
		InformaticaExecutionService informaticaExecutionService = new InformaticaExecutionService(intsvc, informaticaWebServicesHub);
		InformaticaExecutionClient informaticaRepositoryClient = newIntegrationClient(informaticaExecutionService);

		try
		{
			visitor.useClient(informaticaRepositoryClient, runtimeSupport);
		} finally
		{
			cacheLog("Returning Integration service client id " + System.identityHashCode(informaticaRepositoryClient));
			synchronized(integrationServiceClientCache)
			{
				integrationServiceClientCache.get(informaticaExecutionService).returnObject(informaticaRepositoryClient);
			}
		}
	}
}